/*******************************************************************************
Copyright 2016 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/

/** INCLUDES *******************************************************/
#include "usb.h"
#include "usb_device_hid.h"

#include <string.h>

#include "system.h"

#define _XTAL_FREQ 48000000

#define SPIADP_CMD_SPI_MODE		0x10	//SPI[h.
#define SPIADP_CMD_SPI_SPEED	0x11	//SPIx.
#define SPIADP_CMD_ACK_SETTING	0x12	//AckON/OFF.
#define SPIADP_CMD_CS_CTRL		0x20	//CS.
#define SPIADP_CMD_GET_INPUT	0x21	//̓s̏Ԏ擾.
#define SPIADP_CMD_SET_OUTPUT	0x22	//o̓s̏Ԑݒ.
#define SPIADP_CMD_GET_OUTPUT	0x23	//o̓s̏Ԏ擾.
#define SPIADP_CMD_DATA			0x30	//f[^ǂݏ.

#define SPIADP_CMND				0
#define SPIADP_LEN				1
#define SPIADP_DATA				2

#define APIADP_ACK_OFF			0
#define APIADP_ACK_RISING_EDGE	1
#define APIADP_ACK_FALLING_EDGE	2

#define PACKET_SIZE		64
//#define PACKET_SIZE		34

#if 0
////////////////////////////////////////////////////////////////////////////
void serial_putc(unsigned char c)
{
	unsigned char i;

	//X^[grbg.
	SERIAL_TX = 0;
	__delay_us(SERIAL_WAIT);

	for(i = 0; i < 8; i++){
		SERIAL_TX = (c & 0x01);
		__delay_us(SERIAL_WAIT);
		c >>= 1;
	}

	//Xgbvrbg.
	SERIAL_TX = 1;
	__delay_us(SERIAL_WAIT);
}
#endif

////////////////////////////////////////////////////////////////////////////
//void print_word10(unsigned short val)
//{
//	serial_putc((val / 10000) + '0');
//	val %= 10000;
//
//	serial_putc((val / 1000) + '0');
//	val %= 1000;
//
//	serial_putc((val / 100) + '0');
//	val %= 100;
//
//	serial_putc((val / 10) + '0');
//	val %= 10;
//
//	serial_putc(val + '0');
//}

/** VARIABLES ******************************************************/
/* Some processors have a limited range of RAM addresses where the USB module
 * is able to access.  The following section is for those devices.  This section
 * assigns the buffers that need to be used by the USB module into those
 * specific areas.
 */
#if defined(FIXED_ADDRESS_MEMORY)
    #if defined(COMPILER_MPLAB_C18)
        #pragma udata HID_CUSTOM_OUT_DATA_BUFFER = HID_CUSTOM_OUT_DATA_BUFFER_ADDRESS
        //unsigned char ReceivedDataBuffer[64];	//@datasize
        unsigned char ReceivedDataBuffer[PACKET_SIZE];
        #pragma udata HID_CUSTOM_IN_DATA_BUFFER = HID_CUSTOM_IN_DATA_BUFFER_ADDRESS
        //unsigned char ToSendDataBuffer[64];	//@datasize
        unsigned char ToSendDataBuffer[PACKET_SIZE];
        #pragma udata

    #elif defined(__XC8)
        //unsigned char ReceivedDataBuffer[64] HID_CUSTOM_OUT_DATA_BUFFER_ADDRESS;	//@datasize
        unsigned char ReceivedDataBuffer[PACKET_SIZE] HID_CUSTOM_OUT_DATA_BUFFER_ADDRESS;
        //unsigned char ToSendDataBuffer[64] HID_CUSTOM_IN_DATA_BUFFER_ADDRESS;	//@datasize
        unsigned char ToSendDataBuffer[PACKET_SIZE] HID_CUSTOM_IN_DATA_BUFFER_ADDRESS;
    #endif
#else
    //unsigned char ReceivedDataBuffer[64];	//@datasize
    unsigned char ReceivedDataBuffer[PACKET_SIZE];
    //unsigned char ToSendDataBuffer[64];	//@datasize
    unsigned char ToSendDataBuffer[PACKET_SIZE];
#endif

volatile USB_HANDLE USBOutHandle;    
volatile USB_HANDLE USBInHandle;

/** DEFINITIONS ****************************************************/
typedef enum
{
    COMMAND_TOGGLE_LED = 0x80,
    COMMAND_GET_BUTTON_STATUS = 0x81,
    COMMAND_READ_POTENTIOMETER = 0x37
} CUSTOM_HID_DEMO_COMMANDS;

unsigned char bAckFlg = 0;
unsigned char boIntPinFlg = 0;
unsigned char bLedIdleState = 0;

/** FUNCTIONS ******************************************************/

/*********************************************************************
* Function: void APP_DeviceCustomHIDInitialize(void);
*
* Overview: Initializes the Custom HID demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceCustomHIDInitialize()
{
    //initialize the variable holding the handle for the last
    // transmission
    USBInHandle = 0;

    //enable the HID endpoint
    USBEnableEndpoint(CUSTOM_DEVICE_HID_EP, USB_IN_ENABLED|USB_OUT_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);

    //Re-arm the OUT endpoint for the next packet
    //USBOutHandle = (volatile USB_HANDLE)HIDRxPacket(CUSTOM_DEVICE_HID_EP,(uint8_t*)&ReceivedDataBuffer[0],64);	//@datasize
    USBOutHandle = (volatile USB_HANDLE)HIDRxPacket(CUSTOM_DEVICE_HID_EP,(uint8_t*)&ReceivedDataBuffer[0],PACKET_SIZE);	//@datasize
}

////////////////////////////////////////////////////////////////////////////
// J[h1oCgM.
//
// MDOUTf[^Ԃ.
// PɃNbN^AJ[hMꍇɂg.
//
// IN:
//   c Mf[^.
//
// OUT:
//
// RETURN:
//  J[h̃f[^.
//
////////////////////////////////////////////////////////////////////////////
unsigned char MMSendByte(unsigned char c)
{
	unsigned char dummy;
	unsigned short loop = 0;

	boIntPinFlg = 0;

	dummy = SSPBUF;
	SSPBUF = c;
	while(!SSPSTATbits.BF);

	if(bAckFlg == 1){
		while(boIntPinFlg == 0){
			if(loop++ > 10000){		//ő10msec.
				break;
			}
			asm("nop");	//1usec(83.3ns * 12).
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
		}
	}

	return SSPBUF;
}

/*********************************************************************
* Function: void APP_DeviceCustomHIDTasks(void);
*
* Overview: Keeps the Custom HID demo running.
*
* PreCondition: The demo should have been initialized and started via
*   the APP_DeviceCustomHIDInitialize() and APP_DeviceCustomHIDStart() demos
*   respectively.
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceCustomHIDTasks()
{   
    /* If the USB device isn't configured yet, we can't really do anything
     * else since we don't have a host to talk to.  So jump back to the
     * top of the while loop. */
    if( USBGetDeviceState() < CONFIGURED_STATE )
    {
        return;
    }

    /* If we are currently suspended, then we need to see if we need to
     * issue a remote wakeup.  In either case, we shouldn't process any
     * keyboard commands since we aren't currently communicating to the host
     * thus just continue back to the start of the while loop. */
    if( USBIsDeviceSuspended()== true )
    {
        return;
    }
    
    //Check if we have received an OUT data packet from the host
    if(HIDRxHandleBusy(USBOutHandle) == false)
    {   
        //We just received a packet of data from the USB host.
        //Check to make sure the endpoint/buffer is free before we modify the contents
        if (!HIDTxHandleBusy(USBInHandle)) {
			LATCbits.LATC5 = !bLedIdleState;

			memcpy(ToSendDataBuffer, ReceivedDataBuffer, PACKET_SIZE);	//@datasize

			switch(ReceivedDataBuffer[SPIADP_CMND]){
			case SPIADP_CMD_SPI_MODE:	//SPI[h.
				switch(ReceivedDataBuffer[SPIADP_DATA]){
				case 0:
					SSPCON1bits.CKP = 0;	//Mode0
					SSPSTATbits.CKE = 1;	//Mode0
					break;
				case 1:
					SSPCON1bits.CKP = 0;	//Mode1
					SSPSTATbits.CKE = 0;	//Mode1
					break;
				case 2:
					SSPCON1bits.CKP = 1;	//Mode2
					SSPSTATbits.CKE = 1;	//Mode2
					break;
				case 3:
					SSPCON1bits.CKP = 1;	//Mode3
					SSPSTATbits.CKE = 0;	//Mode3
					break;
				}
				break;
			case SPIADP_CMD_SPI_SPEED:	//SPIx.
				if(ReceivedDataBuffer[SPIADP_DATA] == 0){
					SSPCON1bits.SSPM = 0b0000;	//Fosc/4.
					SSPADD = 0;
				}
				else{
					SSPCON1bits.SSPM = 0b1010;	//SSPADDɈˑ.
					SSPADD = ReceivedDataBuffer[SPIADP_DATA];
				}
				break;
			case SPIADP_CMD_ACK_SETTING:	//AckON/OFF.
				switch(ReceivedDataBuffer[SPIADP_DATA]){
				case APIADP_ACK_OFF:
					IOCAPbits.IOCAP4 = 0;		//Ack.
					IOCANbits.IOCAN4 = 0;		//Ack.
					bAckFlg = 0;
					break;
				case APIADP_ACK_RISING_EDGE:
					IOCAPbits.IOCAP4 = 1;		//Ack.
					IOCANbits.IOCAN4 = 0;		//Ack.
					bAckFlg = 1;
					break;
				case APIADP_ACK_FALLING_EDGE:
					IOCAPbits.IOCAP4 = 0;		//Ack.
					IOCANbits.IOCAN4 = 1;		//Ack.
					bAckFlg = 1;
					break;
				}
				break;
			case SPIADP_CMD_CS_CTRL:	//CS.
				switch(ReceivedDataBuffer[SPIADP_DATA]){
				case 0:
					LATCbits.LATC3 = 0;
					break;
				case 1:
					LATCbits.LATC3 = 1;
					break;
				}
				break;
			case SPIADP_CMD_GET_INPUT:	//̓s̏Ԏ擾.
				ToSendDataBuffer[SPIADP_DATA] = RA4;
				ToSendDataBuffer[SPIADP_LEN] = 1;
				break;
			case SPIADP_CMD_SET_OUTPUT:	//o̓s̏Ԑݒ.
				switch(ReceivedDataBuffer[SPIADP_DATA]){
				case 0:
					LATCbits.LATC4 = 0;
					break;
				case 1:
					LATCbits.LATC4 = 1;
					break;
				}
				break;
			case SPIADP_CMD_GET_OUTPUT:	//o̓s̏Ԏ擾.
				ToSendDataBuffer[SPIADP_DATA] = RC4;
				ToSendDataBuffer[SPIADP_LEN] = 1;
				break;
			case 0x30:	//f[^ǂݏ.
				for(unsigned char i = 0; i < ReceivedDataBuffer[SPIADP_LEN]; i++){
					ToSendDataBuffer[SPIADP_DATA + i] = MMSendByte(ReceivedDataBuffer[SPIADP_DATA + i]);
				}
				break;
			default:
				break;
			}

            USBInHandle = HIDTxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*) & ToSendDataBuffer[0], PACKET_SIZE);	//@datasize

			LATCbits.LATC5 = bLedIdleState;
        }
        //Re-arm the OUT endpoint, so we can receive the next OUT data packet 
        //that the host may try to send us.
        USBOutHandle = HIDRxPacket(CUSTOM_DEVICE_HID_EP, (uint8_t*)&ReceivedDataBuffer[0], PACKET_SIZE);	//@datasize
    }
}